from __future__ import annotations

from copy import copy

import numpy as np
import xarray as xr

from skretrieval.retrieval.statevector import StateVectorElement
from skretrieval.retrieval.tikhonov import (
    two_dim_vertical_first_deriv,
)


class WavenumberShift(StateVectorElement):
    def __init__(
        self,
        num_los: int,
        tikh_factor: float,
        prior_factor: float = 0,
        numerical_delta=0.0001,
        min_shift=-0.1,
        max_shift=0.1,
        apply_to_measurement: str | None = None,
    ):
        """
        Implements a wavenumber shift for every modelled LOS in the forward model

        Parameters
        ----------
        num_los : int
        numerical_delta : float, optional
            _description_, by default 0.0001
        """
        self._shifts = np.zeros(num_los)
        self._numerical_delta = numerical_delta
        self._tikh_factor = tikh_factor
        self._prior_factor = prior_factor

        self._min_shift = min_shift
        self._max_shift = max_shift
        self._apply_to_measurement = apply_to_measurement

        super().__init__(True)

    def state(self) -> np.array:
        return copy(self._shifts)

    def name(self) -> str:
        if self._apply_to_measurement:
            return f"wavenumber_shifts_{self._apply_to_measurement}"
        return "wavenumber_shifts"

    def lower_bound(self) -> np.array:
        return np.ones_like(self._shifts.flatten()) * self._min_shift

    def upper_bound(self) -> np.array:
        return np.ones_like(self._shifts.flatten()) * self._max_shift

    def inverse_apriori_covariance(self) -> np.ndarray:
        gamma = two_dim_vertical_first_deriv(
            1, len(self._shifts), factor=self._tikh_factor
        )
        return gamma.T @ gamma + np.eye(len(self._shifts)) * self._prior_factor

    def apriori_state(self) -> np.array:
        return np.zeros_like(self.state())

    def propagate_wf(self, radiance) -> np.ndarray:
        wf = np.zeros(
            (
                len(self._shifts),
                radiance["radiance"].shape[0],
                radiance["radiance"].shape[1],
                radiance["radiance"].shape[2],
            )
        )

        for i in range(len(self._shifts)):
            new_rad = radiance.isel(los=i, stokes=0)["radiance"].interp(
                wavelength=radiance.wavelength + self._numerical_delta,
                kwargs={"fill_value": "extrapolate"},
            )

            drad = (
                new_rad - radiance.isel(los=i, stokes=0)["radiance"]
            ) / self._numerical_delta
            wf[i, :, i, 0] = drad
        return xr.DataArray(wf, dims=["x", "wavelength", "los", "stokes"])

    def update_state(self, x: np.array):
        self._shifts = copy(x)

    def modify_input_radiance(self, radiance: xr.Dataset):
        for i in range(len(self._shifts)):
            shift = self._shifts[i]
            for var in list(radiance):
                if "wavelength" in radiance[var].dims:
                    if len(radiance[var].dims) == 3:
                        radiance[var].to_numpy()[:, i, 0] = radiance.isel(
                            los=i, stokes=0
                        )[var].interp(
                            wavelength=radiance.wavelength + shift,
                            kwargs={"fill_value": "extrapolate"},
                        )
                    else:
                        radiance[var].to_numpy()[:, :, i, 0] = radiance.isel(
                            los=i, stokes=0
                        )[var].interp(
                            wavelength=radiance.wavelength + shift,
                            kwargs={"fill_value": "extrapolate"},
                        )

        return radiance



from __future__ import annotations

import abc
from collections.abc import Callable

import numpy as np
import sasktran2 as sk

from skretrieval.core.lineshape import DeltaFunction, LineShape
from skretrieval.core.sasktranformat import SASKTRANRadiance
from skretrieval.core.sensor.spectrograph import SpectrographOnlySpectral
from skretrieval.geodetic import geodetic
from skretrieval.retrieval import ForwardModel

from .ancillary import Ancillary
from .measvec import MeasurementVector
from .observation import FilteredObservation, Observation
from .statevector.altitude import AltitudeNativeStateVector


class StandardForwardModel(ForwardModel):
    def __init__(
        self,
        observation: Observation,
        state_vector: AltitudeNativeStateVector,
        meas_vec: MeasurementVector,
        ancillary: Ancillary,
        engine_config: sk.Config,
        **kwargs,
    ) -> None:
        """
        A forward model for the Retrieval class.  This is a base class that should be inherited from.


        Parameters
        ----------
        observation : Observation
            The observation
        state_vector : AltitudeNativeStateVector
            The State Vector
        ancillary : Ancillary
            The Ancillary Object
        engine_config : sk.Config
            Configuration for the engine
        """
        ForwardModel.__init__(self)

        self._state_vector = state_vector
        self._engine_config = engine_config
        self._ancillary = ancillary
        self._observation = observation
        self._meas_vec = meas_vec

        self._viewing_geo = self._construct_viewing_geo()
        self._model_geometry = self._construct_model_geometry()
        self._model_wavelength = self._construct_model_wavelength()

        self._atmosphere = self._construct_atmosphere()

        self._engine = self._construct_engine()

        self._inst_model = self._construct_inst_model()

        self._solar_model = self._construct_solar_model()

    @abc.abstractmethod
    def _construct_model_geometry(self):
        pass

    @abc.abstractmethod
    def _construct_model_wavelength(self):
        pass

    @abc.abstractmethod
    def _construct_viewing_geo(self):
        pass

    @abc.abstractmethod
    def _construct_inst_model(self):
        pass

    def _construct_solar_model(self):
        pass

    def _construct_atmosphere(self):
        atmo = {}

        for key in self._model_geometry:
            atmo[key] = sk.Atmosphere(
                self._model_geometry[key],
                self._engine_config,
                wavelengths_nm=self._model_wavelength[key],
                pressure_derivative=False,
                temperature_derivative=False,
            )

            self._state_vector.add_to_atmosphere(atmo[key])
            self._ancillary.add_to_atmosphere(atmo[key])

            self._model_geometry[
                key
            ].refractive_index = sk.optical.refraction.ciddor_index_of_refraction(
                atmo[key].temperature_k,
                atmo[key].pressure_pa,
                np.zeros_like(atmo[key].temperature_k),
                450,
                np.nanmean(atmo[key].wavelengths_nm),
            )

        return atmo

    def _construct_engine(self):
        engines = {}

        for key in self._model_geometry:
            engines[key] = sk.Engine(
                self._engine_config, self._model_geometry[key], self._viewing_geo[key]
            )

        return engines

    def calculate_radiance(self):
        l1 = {}
        for key in self._engine:
            sk2_rad = self._engine[key].calculate_radiance(self._atmosphere[key])
            sk2_rad = self._state_vector.post_process_sk2_radiances(sk2_rad)
            sk2_rad = SASKTRANRadiance.from_sasktran2(sk2_rad)

            model_result = self._inst_model[key].model_radiance(sk2_rad, None)

            if isinstance(model_result, dict):
                if len(model_result) == 1:
                    l1[key] = next(iter(model_result.values()))
                else:
                    for k, v in model_result.items():
                        l1[f"{key}_{k}"] = v
            else:
                l1[key] = model_result

            self._observation.append_information_to_l1(l1)

        return l1


class SpectrometerMixin:
    def __init__(
        self,
        lineshape_fn: Callable[[float], LineShape] | None = None,
        model_res_nm=0.02,
        model_res_cminv=0.02,
        spectral_native_coordinate="wavelength_nm",
        round_decimal=2,
        stokes_sensitivities=None,
    ) -> None:
        """
        Mixin for adding a spectrometer to the forward model

        Parameters
        ----------
        lineshape_fn : Callable[[float], LineShape] | None, optional
            Function that takes in wavelength in nm and returns back a LineShape, by default None
        model_res_nm : float, optional
            Model Resolution to use in [nm], by default 0.02
        model_res_cminv : float, optional
            Model Resolution to use in [cm^-1], by default 0.02
        spectral_native_coordinate : str, optional
            The native coordinate for the spectral axis, by default "wavelength_nm", can also be
            "wavenumber_cminv"
        round_decimal : int, optional
            Decimal points to round the wavelengths to in the radiative transfer calculation, by default 2
        stokes_sensitivities : dict, optional
            Dictionary of stokes sensitivities, by default None. Can be set to multiple measurements, e.g.,
            {"I": np.array([1, 0, 0, 0]), "Q": np.array([0, 1, 0, 0]), "U": np.array([0, 0, 1, 0])} to measure
            multiple stokes parameters separately.
        """
        if lineshape_fn is None:
            self._lineshape_fn = lambda _: DeltaFunction()
        else:
            self._lineshape_fn = lineshape_fn

        self._model_res_nm = model_res_nm
        self._model_res_cminv = model_res_cminv
        self._round_decimal = round_decimal
        self._spectral_native_coordinate = spectral_native_coordinate
        self._stokes_sensitivities = stokes_sensitivities

    def _get_required_wavelength(self):
        obs_samples = self._observation.sample_wavelengths()
        mv_required_samples = {}
        for key, val in self._meas_vec.items():
            mv_required_samples[key] = val.required_sample_wavelengths(obs_samples)

        sample_wavelengths = {}
        for key in obs_samples:
            sample_wavelengths[key] = np.unique(
                np.concatenate(
                    [np.atleast_1d(d[key]) for d in mv_required_samples.values()]
                )
            )

        return sample_wavelengths

    def _construct_model_wavelength(self):
        """
        Evaluates the lineshape at the sample wavelengths and returns back the model wavelengths
        spaced by the model resolution
        """
        sample_wavelengths = self._get_required_wavelength()

        ws = {}
        for k, v in sample_wavelengths.items():
            if self._spectral_native_coordinate == "wavelength_nm":
                bounds = [
                    self._lineshape_fn(w).bounds(center=0)
                    if self._lineshape_fn(w).zero_centered()
                    else self._lineshape_fn(w).bounds(center=-w)
                    for w in v
                ]
                ws[k] = np.unique(
                    np.concatenate(
                        [
                            np.around(
                                np.arange(
                                    a, b + self._model_res_nm / 2, self._model_res_nm
                                )
                                + w,
                                self._round_decimal,
                            )
                            for (a, b), w in zip(bounds, v)
                        ]
                    )
                )
            else:
                bounds = [
                    self._lineshape_fn(w).bounds(center=0)
                    if self._lineshape_fn(w).zero_centered()
                    else self._lineshape_fn(w).bounds(center=-1e7 / w)
                    for w in v
                ]
                ws[k] = (
                    1e7
                    / np.unique(
                        np.concatenate(
                            [
                                np.around(
                                    np.arange(
                                        1e7 / (b + w),
                                        1e7 / (a + w) + self._model_res_cminv / 2,
                                        self._model_res_cminv,
                                    ),
                                    self._round_decimal,
                                )
                                for (a, b), w in zip(bounds, v)
                            ]
                        )
                    )[::-1]
                )

        return ws

    def _construct_inst_model(self):
        """
        Constructs the instrument model
        """
        sample_wavelengths = self._get_required_wavelength()

        inst_models = {}

        for key in sample_wavelengths:
            inst_models[key] = SpectrographOnlySpectral(
                sample_wavelengths[key],
                [self._lineshape_fn(x) for x in sample_wavelengths[key]],
                spectral_native_coordinate=self._spectral_native_coordinate,
                assign_coord="wavelength"
                if self._spectral_native_coordinate == "wavelength_nm"
                else "wavenumber",
                stokes_sensitivity=self._stokes_sensitivities,
            )

        return inst_models


class IdealViewingMixin:
    def __init__(self, observation: Observation, model_altitude_grid: np.array) -> None:
        """
        Mixin for adding an ideal viewing geometry to the forward model. This means
        that a single line of sight is used by the forward model for each observation rather
        than using a spatial PSF.

        Parameters
        ----------
        observation : Observation
        model_altitude_grid : np.array
        """
        self._model_altitude_grid = model_altitude_grid
        self._obs = observation

    def _construct_viewing_geo(self):
        return self._obs.sk2_geometry()

    def _construct_model_geometry(self):
        # Construct the model geometry

        # State vector tells us the engine altitude grid
        altitude_grid_m = self._model_altitude_grid

        # Observation tells us the reference point
        cos_sza = self._obs.reference_cos_sza()
        ref_lat = self._obs.reference_latitude()
        ref_lon = self._obs.reference_longitude()

        geometry = {}

        geo = geodetic()
        for key in cos_sza:
            geo.from_lat_lon_alt(ref_lat[key], ref_lon[key], 0.0)
            earth_radius_m = np.linalg.norm(geo.location)
            geometry[key] = sk.Geometry1D(
                cos_sza[key],
                0.0,
                earth_radius_m,
                altitude_grid_m,
                sk.InterpolationMethod.LinearInterpolation,
                sk.GeometryType.Spherical,
            )

        return geometry


class IdealViewingSpectrograph(
    IdealViewingMixin, SpectrometerMixin, StandardForwardModel
):
    def __init__(
        self,
        observation: Observation,
        state_vector: AltitudeNativeStateVector,
        meas_vec: MeasurementVector,
        ancillary: Ancillary,
        engine_config: sk.Config,
        **kwargs,
    ) -> None:
        """
        A forward model for the retrieval that uses an ideal viewing geometry and a spectrometer

        Parameters
        ----------
        observation : Observation
        state_vector : AltitudeNativeStateVector
        ancillary : Ancillary
        engine_config : sk.Config
        kwargs : dict
            Additional arguments passed to `SpectrometerMixin`
        """
        IdealViewingMixin.__init__(self, observation, state_vector.altitude_grid)
        SpectrometerMixin.__init__(
            self,
            lineshape_fn=kwargs.get("lineshape_fn", lambda _: DeltaFunction()),
            model_res_cminv=kwargs.get("model_res_cminv", 0.02),
            model_res_nm=kwargs.get("model_res_nm", 0.02),
            spectral_native_coordinate=kwargs.get(
                "spectral_native_coordinate", "wavelength_nm"
            ),
            stokes_sensitivities=kwargs.get("stokes_sensitivities"),
        )
        StandardForwardModel.__init__(
            self,
            observation,
            state_vector,
            meas_vec,
            ancillary,
            engine_config,
            **kwargs,
        )


class ForwardModelHandler(ForwardModel):
    def __init__(
        self,
        cfg: dict,
        observation: Observation,
        state_vector: AltitudeNativeStateVector,
        meas_vec: MeasurementVector,
        ancillary: Ancillary,
        engine_config: sk.Config,
        **kwargs,
    ):
        super().__init__()

        # Construct the internal forward models
        self._forward_models = {}
        for key, val in cfg.items():
            self._forward_models[key] = val.get("class", IdealViewingSpectrograph)(
                FilteredObservation(observation, key),
                state_vector,
                meas_vec,
                ancillary,
                engine_config,
                **val.get("kwargs", {}),
            )

    def calculate_radiance(self):
        result = {}
        for _, v in self._forward_models.items():
            result.update(v.calculate_radiance())
        return result



from __future__ import annotations

import numpy as np
from scipy.sparse import lil_matrix


def two_dim_vertical_second_deriv(numangle, numalt, factor=1, sparse=False):
    """
    Calculates the second derivatvie Tikhonov regularization matrix for a two dimensional uniform grid.
    The matrix is calculated assuming that the measurement vector is constructed with altitude being the leading
    dimension

    Parameters
    ----------
    numangle : scalar
        The number of angular grid points
    numalt : scalar
        The number of altitude grid points
    factor : scalar or length numalt, optional
        If scalar, the resulting matrix is multiplied by this value.  If a vector of length numalt, then each altitude
        level is multiplied by its corresponding factor.  Default is 1
    """

    n = numalt * numangle
    gamma = lil_matrix((n, n)) if sparse else np.zeros((n, n))
    for idangle in range(numangle):
        for idalt in range(1, numalt - 1):
            mfactor = factor[idalt] if np.shape(factor) != () else factor
            idx = idangle * numalt + idalt
            gamma[idx, idx - 1] = -1 / 4 * mfactor
            gamma[idx, idx] = 1 / 2 * mfactor
            gamma[idx, idx + 1] = -1 / 4 * mfactor

    if sparse:
        gamma = gamma.asformat("csr")

    return gamma


def two_dim_vertical_first_deriv(numangle, numalt, factor=1, sparse=False):
    """
    Calculates the first derivatvie Tikhonov regularization matrix for a two dimensional uniform grid.
    The matrix is calculated assuming that the measurement vector is constructed with altitude being the leading
    dimension

    Parameters
    ----------
    numangle : scalar
        The number of angular grid points
    numalt : scalar
        The number of altitude grid points
    factor : scalar or length numalt, optional
        If scalar, the resulting matrix is multiplied by this value.  If a vector of length numalt, then each altitude
        level is multiplied by its corresponding factor.  Default is 1
    """

    n = numalt * numangle
    gamma = lil_matrix((n, n)) if sparse else np.zeros((n, n))
    for idangle in range(numangle):
        for idalt in range(numalt - 1):
            mfactor = factor[idalt] if np.shape(factor) != () else factor
            idx = idangle * numalt + idalt
            gamma[idx, idx] = -1 / 2 * mfactor
            gamma[idx, idx + 1] = 1 / 2 * mfactor

    if sparse:
        gamma = gamma.asformat("csr")

    return gamma


def two_dim_horizontal_second_deriv(numangle, numalt, factor=1, sparse=False):
    """
    Calculates the second derivatvie Tikhonov regularization matrix for a two dimensional uniform grid.
    The matrix is calculated assuming that the measurement vector is constructed with altitude being the leading
    dimension

    Parameters
    ----------
    numangle : scalar
        The number of angular grid points
    numalt : scalar
        The number of altitude grid points
    factor : scalar or vector length numangle, optional
        If scalar, the resulting matrix is multiplied by this value.  If a vector of length numangle, then each angular
        level is multiplied by the corresponding value. Default is 1
    """

    n = numalt * numangle
    gamma = lil_matrix((n, n)) if sparse else np.zeros((n, n))
    for idx in range(numalt, n - numalt):
        if np.shape(factor) != ():
            horiz = int(idx / numalt)
            mfactor = factor[horiz]
        else:
            mfactor = factor
        gamma[idx, idx - numalt] = -1 / 4 * mfactor
        gamma[idx, idx] = 1 / 2 * mfactor
        gamma[idx, idx + numalt] = -1 / 4 * mfactor

    if sparse:
        # Convert gamma to CSR format for faster matrix operations
        gamma = gamma.asformat("csr")
    return gamma



from __future__ import annotations

from collections import namedtuple

# Quantities that are necessary to uniquely define the measurement geometry for a specified instrument
OpticalGeometry = namedtuple(
    "OpticalGeometry", ["observer", "look_vector", "local_up", "mjd"]
)



from __future__ import annotations

from skretrieval.core.radianceformat import RadianceGridded
from skretrieval.retrieval.measvec import (
    MeasurementVector,
    concat,
    post_process,
    pre_process,
)
from skretrieval.retrieval.statevector import StateVector
from skretrieval.retrieval.target import GenericTarget


class MeasVecTarget(GenericTarget):
    def __init__(
        self,
        state_vector: StateVector,
        measurement_vectors: dict[MeasurementVector],
        context: dict,
        rescale_state_space: bool = False,
    ):
        """
        A target where the measurement vector is calculated through MeasurementVector objects

        Parameters
        ----------
        state_vector : StateVector
        measurement_vectors : dict[MeasurementVector]
        rescale_state_space : bool, optional
            If true, the state vectors are internally scaled to be within their min and max values, by default False
        """
        super().__init__(state_vector, rescale_state_elements=rescale_state_space)
        self._measurement_vectors = measurement_vectors
        self._context = context

    def _internal_measurement_vector(self, l1_data: dict[RadianceGridded]):
        l1 = pre_process(l1_data, len(self.state_vector()))

        res = []
        for _, v in self._measurement_vectors.items():
            appl = v.apply(l1, self._context)
            if appl is not None:
                res.append(appl)

        return post_process(concat(res))



from __future__ import annotations

from copy import copy

import numpy as np
import sasktran as sk
import xarray as xr

from skretrieval.legacy.retrieval.statevector.profile import StateVectorProfile


class AerosolGaussianPressure(StateVectorProfile):
    def __init__(
        self,
        clim_alts: np.array,
        vertical_aod: float,
        aod_wavelength: float,
        layer_height: float,
        layer_width: float,
        species_name: str,
        optical_property: sk.OpticalProperty,
        prior_aod: float,
        prior_layer_height: float,
        prior_layer_width: float,
        prior_aod_sigma: float,
        prior_layer_height_sigma: float,
        prior_layer_width_sigma: float,
        background_climatology: sk.ClimatologyUserDefined,
    ):
        """
        Component of the state vector that is an arbitrary aerosol type with a vmr profile that
        has a Gaussian shape in pressure space.

        The state vector contains 3 elements, the vertical aod at a reference wavelength, the layer height, and
        the layer width.  The layer height and layer width are specified with a vertical coordinate of
        x = pressure / (surface pressure)


        Parameters
        ----------
        clim_alts: np.array
            Internal climatology altitudes to represent the profile on
        vertical_aod: float
            Vertical optical depth at aod_wavelength
        aod_wavelength: float
            Wavelength in nm that the vertical_aod is specified at
        layer_height: float
            Central layer height in units of pressure / (surface pressure)
        layer_width: float
            Gaussian 1sigma width of the layer in units of pressure / (surface pressure)
        species_name: str
            Name of the species
        optical_property: sk.OpticalProperty
            Optical property of the species
        prior_aod: float
            Prior value for the AOD
        prior_layer_height: float
            Prior value for the layer height
        prior_layer_width: float
            Prior value for the layer width
        prior_aod_sigma: float
            Prior 1sigma on the vertical aod
        prior_layer_height_sigma: float
            Prior 1sigma on the layer height
        prior_layer_width_sigma: float
            Prior 1sigma on the layer width
        background_climatology: sk.Climatology
            Climatology containing 'SKCLIMATOLOGY_PRESSURE_PA' and 'SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3'
        """
        self._background_climatology = background_climatology
        self._background_pressure = np.asarray(
            self._background_climatology["SKCLIMATOLOGY_PRESSURE_PA"]
        )
        self._background_numberdensity = np.asarray(
            self._background_climatology["SKCLIMATOLOGY_AIRNUMBERDENSITY_CM3"]
        )

        # State params are vertical AOD, layer height, and layer width
        self._clim_alts = clim_alts
        self._vertical_aod = vertical_aod
        self._layer_height = layer_height
        self._layer_width = layer_width
        self._aod_wavelength = aod_wavelength

        self._prior_aod = prior_aod
        self._prior_layer_height = prior_layer_height
        self._prior_layer_width = prior_layer_width
        self._prior_aod_sigma = prior_aod_sigma
        self._prior_layer_height_sigma = prior_layer_height_sigma
        self._prior_layer_width_sigma = prior_layer_width_sigma

        self._xs = optical_property.calculate_cross_sections(
            sk.MSIS90(), 0, 0, 10000, 54372, [self._aod_wavelength]
        ).total[0]

        super().__init__(species_name, optical_property)

    def inverse_apriori_covariance(self) -> np.ndarray:
        n = len(self.state())
        inv_apriori_covariance = np.zeros((n, n))

        inv_apriori_covariance[0, 0] = 1 / (self._prior_aod_sigma**2)
        inv_apriori_covariance[1, 1] = 1 / (self._prior_layer_height_sigma**2)
        inv_apriori_covariance[2, 2] = 1 / (self._prior_layer_width_sigma**2)

        return inv_apriori_covariance

    def apriori_state(self) -> np.array:
        return np.array(
            [self._prior_aod, self._prior_layer_height, self._prior_layer_width]
        )

    def _climatology_altitudes_and_density(self):
        return self._clim_alts, self._nd(
            self._vertical_aod, self._layer_height, self._layer_width
        )

    def _nd(self, aod, layer_height, layer_width):
        # vmr gaussian
        gauss = np.exp(
            -0.5
            * (
                (
                    self._background_pressure / self._background_pressure[0]
                    - layer_height
                )
                / (2 * layer_width)
            )
            ** 2
        )

        # convert to number density space
        gauss *= self._background_numberdensity

        # Convert profile to have the correct vertical aod
        integral = np.trapz(gauss, self._clim_alts)

        # xs in cm2, integral is over m to we have to divide by 100
        nd = gauss / integral * aod / self._xs / 100

        return copy(nd)

    def state(self) -> np.array:
        return np.array(
            [
                copy(self._vertical_aod),
                copy(self._layer_height),
                copy(self._layer_width),
            ]
        )

    def propagate_wf(self, radiance) -> np.ndarray:
        num_factor = 1.01

        d_aod = self._nd(
            self._vertical_aod * num_factor, self._layer_height, self._layer_width
        )
        d_height = self._nd(
            self._vertical_aod, self._layer_height * num_factor, self._layer_width
        )
        d_width = self._nd(
            self._vertical_aod, self._layer_height, self._layer_width * num_factor
        )

        base = self._nd(self._vertical_aod, self._layer_height, self._layer_width)

        d_aod = (d_aod - base) / (self._vertical_aod * (num_factor - 1))
        d_height = (d_height - base) / (self._layer_height * (num_factor - 1))
        d_width = (d_width - base) / (self._layer_width * (num_factor - 1))

        matrix = np.vstack([d_aod, d_height, d_width])

        old_wf = radiance["wf_" + self.name()]

        return old_wf @ xr.DataArray(
            data=matrix.T,
            dims=["perturbation", "x"],
            coords={"perturbation": old_wf["perturbation"].to_numpy()},
        )

    def update_state(self, x: np.array):
        self._vertical_aod = copy(x[0])
        self._layer_height = copy(x[1])
        self._layer_width = copy(x[2])
        self._internal_update()



from __future__ import annotations

from collections.abc import Callable
from copy import copy

import numpy as np
import sasktran as sk
import sasktranif.sasktranif as skif
from sasktran.exceptions import wrap_skif_functionfail

from skretrieval.core.lineshape import LineShape


class OpticalPropertyLineShape(sk.OpticalProperty):
    """
    An optical property that is convolved with an arbitrary line shape
    """

    def __init__(
        self,
        optical_property: sk.OpticalProperty,
        line_shape_fn: Callable[[float], LineShape],
        wavel_to_calc_for=None,
    ):
        super().__init__(optical_property._name)

        self._unconv_optical_prop = optical_property
        self._line_shape_fn = line_shape_fn
        self._valid_wavelengths = []
        self._wavel_to_calc_for = wavel_to_calc_for

        if self._wavel_to_calc_for is not None:
            self._iskopticalproperty = self.convolved_optical_property(
                self._unconv_optical_prop, self._line_shape_fn, self._wavel_to_calc_for
            )

    def skif_object(self, **kwargs):
        if len(self._valid_wavelengths) != len(
            kwargs["engine"].wavelengths
        ) or not np.allclose(self._valid_wavelengths, kwargs["engine"].wavelengths):
            self._valid_wavelengths = kwargs["engine"].wavelengths

            calc_wavelengths = copy(self._valid_wavelengths)
            calc_wavelengths = np.concatenate(
                (
                    [np.nanmin(calc_wavelengths) - 1],
                    calc_wavelengths,
                    [np.nanmax(calc_wavelengths) + 1],
                )
            )

            self._iskopticalproperty = self.convolved_optical_property(
                self._unconv_optical_prop, self._line_shape_fn, calc_wavelengths
            )
        return self._iskopticalproperty

    def __setstate__(self, state):
        super().__setstate__(state)
        self._iskopticalproperty = self.convolved_optical_property(
            self._unconv_optical_prop, self._line_shape_fn, self._wavel_to_calc_for
        )

    def __repr__(self):
        return f"SasktranIF Optical Property: {self._name}_Convolved"

    @staticmethod
    @wrap_skif_functionfail
    def convolved_optical_property(
        hires_optprop: sk.OpticalProperty,
        line_shape_fn: Callable[[float], LineShape],
        wavel_to_calc_for,
    ):
        """
        Convolves down the hi-resolution cross sections and creates a user defined optical property. Convolution assumes
        the hi-resolution version has infinite resolution

        Parameters
        ----------
        hires_optprop : sasktran.OpticalProperty
            sasktran OpticalProperty that will be convolved down to desired resolution.
        Returns
        -------
        optprop : ISKOpticalProperty
            Convolved optical property
        """

        import contextlib

        optprop = skif.ISKOpticalProperty("USERDEFINED_TABLES")
        with contextlib.suppress(skif._sasktranif.functionfail):
            optprop.SetProperty("WavelengthTruncation", 1)

        # Temperatures of the hi-resolution xsect
        temp = hires_optprop.info["temperatures"]
        # Ranges in which each temperature has data
        temp_range = hires_optprop.info["wavelength_range"]

        spec_sampling = hires_optprop.info["spectral sampling"] = 0.01

        for T, ra in zip(temp, temp_range):
            # Get the high resolution DBM cross section
            clim_temp = sk.ClimatologyUserDefined(
                [0, 10000], {"SKCLIMATOLOGY_TEMPERATURE_K": [T, T]}, interp="linear"
            )
            wavel_hr = np.arange(ra[0], ra[1], spec_sampling)

            if wavel_to_calc_for is None:
                wavel_calc = wavel_hr
            else:
                wavel_calc = wavel_to_calc_for
                wavel_calc = wavel_calc[(wavel_calc >= ra[0]) & (wavel_calc <= ra[1])]

            xsect = hires_optprop.calculate_cross_sections(
                clim_temp, 0, 0, 1000, 54372, wavel_hr
            )
            xsect = xsect.absorption

            wavelconv = np.zeros(np.shape(wavel_hr))
            xsectconv = np.zeros(np.shape(wavel_hr))

            for widx, w in enumerate(wavel_calc):
                line_shape = line_shape_fn(w)
                weights = line_shape.integration_weights(w, wavel_hr)
                wavelconv[widx] = np.dot(wavel_hr, weights)
                xsectconv[widx] = np.dot(xsect, weights)

            good = xsectconv > 0
            if np.nansum(good) > 0:
                full_xsect = np.interp(
                    wavel_hr,
                    wavelconv[good],
                    xsectconv[good],
                    left=np.nan,
                    right=np.nan,
                )

                optprop.AddUserDefined(
                    T,
                    wavel_hr[~np.isnan(full_xsect)],
                    full_xsect[~np.isnan(full_xsect)],
                )

        return optprop



from __future__ import annotations

import numpy as np
import sasktran2 as sk2

from skretrieval.retrieval.statevector.constituent import StateVectorElementConstituent


def test_statevector_constituent_vmr():
    const = sk2.climatology.mipas.constituent("O3", sk2.optical.O3DBM())

    sv = StateVectorElementConstituent(const, "O3", ["vmr"])

    assert sv.state().shape == (50,)


def test_statevector_constituent_multiple_params():
    alt_grid = np.arange(0, 100001, 1000.0)

    const = sk2.test_util.scenarios.test_aerosol_constituent(alt_grid)

    sv = StateVectorElementConstituent(
        const, "Aerosol", ["extinction_per_m", "lognormal_median_radius"]
    )

    assert sv.state().shape == (202,)



from __future__ import annotations

import numpy as np
import sasktran2 as sk

import skretrieval.retrieval.prior as prior
from skretrieval.retrieval.statevector.constituent import StateVectorElementConstituent


def _sv_ele():
    # dummy constituent for testing
    const = sk.climatology.mipas.constituent("o3", sk.optical.O3DBM())

    return StateVectorElementConstituent(const, "o3", property_names=["vmr"])


def test_vert_tikh():
    alt = np.arange(0, 50000, 500)
    for order in range(1, 3):
        _ = prior.VerticalTikhonov(alt, order)


def test_multiplicative_prior():
    one = 4 * prior.VerticalTikhonov(1)

    two = prior.VerticalTikhonov(1)

    sv = _sv_ele()

    one.init(sv)
    two.init(sv)

    np.testing.assert_allclose(one.inverse_covariance, two.inverse_covariance * 4)


def test_additive_prior():
    one = 4 * prior.VerticalTikhonov(1)

    two = prior.VerticalTikhonov(1)

    p = one + two

    p.init(_sv_ele())

    _ = p.state
    _ = p.inverse_covariance



from __future__ import annotations

from skretrieval import plotting
from skretrieval.core import lineshape
from skretrieval.retrieval import forwardmodel, measvec, observation, statevector
from skretrieval.retrieval.processing import Retrieval



from __future__ import annotations

import numpy as np
from sasktran import geometry

from skretrieval.core import OpticalGeometry


class ScanningSatellite:
    def __init__(
        self,
        instrument,
        engine,
    ):
        self._engine = engine
        self._instrument = instrument


def fake_satellite_scan(altitudes=None):
    geo = geometry.VerticalImage()

    if altitudes is None:
        altitudes = np.arange(25, 40, 0.1)

    geo.from_sza_saa(60, 60, 0, 0, altitudes, 54372, 0)

    optical_axis = []

    for los in geo.lines_of_sight:
        local_up = los.observer / np.linalg.norm(los.observer)

        optical_axis.append(
            OpticalGeometry(
                look_vector=los.look_vector,
                observer=los.observer,
                mjd=los.mjd,
                local_up=local_up,
            )
        )

    return optical_axis, geo.sun



from __future__ import annotations

import numpy as np
from sasktran2.geodetic import WGS84


def geodetic() -> WGS84:
    return WGS84()


def target_lat_lon_alt(
    los_vector: np.array, obs_position: np.array
) -> tuple[float, float, float]:
    """
    Parameters
    ----------
    los_vector : np.array
        3 Element look vector in ECI
    obs_position : np.array
        3 element observer position in ECI

    Returns
    -------
    Tuple[float, float, float]
        Tangent latitude/longitude/altitude if limb looking. Ground latitude/longitude if nadir looking
    """
    location = geodetic()

    location.from_tangent_point(obs_position, los_vector)

    if location.altitude > 0:
        return location.latitude, location.longitude, location.altitude

    location.from_xyz(location.altitude_intercepts(0, obs_position, los_vector)[0])

    return location.latitude, location.longitude, location.altitude



from __future__ import annotations

import xarray as xr


class SASKTRANRadiance:
    def __init__(self, ds: xr.Dataset, collapse_scalar_stokes: bool = True) -> None:
        """
        Handles conversions between Sasktran2 and Sasktran legacy radiance formats, as well as potentially
        other RTM radiance output formats.

        The base SASKTRAN radiance container contains the following variables:

        - radiance: containing dimensions [stokes, los, spectral_grid]

        And the following coordinates:

        - wavelength_nm: wavelength in nm, corresponding to the spectral_grid dimension
        - wavenumber_cminv: wavenumber in cm^-1, corresponding to the spectral_grid dimension

        The "stokes" dimension can either be scalar, or contain any number of the parameters
        ["I", "Q", "U", "V"] for polarized calculations

        These are the minimum required variables and coordinates, others may be added in as needed.

        Weighting functions are included in the dataset as separate variables prefixed by wf_.
        For example wf_ozone_vmr.  These variables must have the dimensions

        - wf: [stokes, los, spectral_grid, ...]

        where the ... may be any number of dimensions and could be different for each wf.

        If collapse_scalar_stokes is set to true, upon construction, the Stokes dimension will be collapsed
        if it is 1 element long.  This is the default behavior.

        Parameters
        ----------
        ds : xr.Dataset

        collapse_scalar_stokes : bool
            If true, then if a scalar calculation is performed the stokes dimension will be removed. By default True
        """
        self._data = ds

        self._validate_ds()

        if collapse_scalar_stokes and self.data.stokes.size == 1:
            self._data = self.data.squeeze("stokes")

    @property
    def data(self) -> xr.Dataset:
        return self._data

    @classmethod
    def from_sasktran_legacy_xr(
        cls, sasktran_legacy_radiance: xr.Dataset
    ) -> SASKTRANRadiance:
        # Check if we have polarized radiances or unpolarized
        if "radiance" in sasktran_legacy_radiance:
            # Scalar
            sasktran_legacy_radiance["radiance"] = sasktran_legacy_radiance[
                "radiance"
            ].expand_dims("stokes")

            for var in list(sasktran_legacy_radiance.variables):
                if var.startswith("wf_"):
                    sasktran_legacy_radiance[var] = sasktran_legacy_radiance[
                        var
                    ].expand_dims("stokes")
        else:
            # Polarized
            sasktran_legacy_radiance["radiance"] = xr.concat(
                [
                    sasktran_legacy_radiance["I"],
                    sasktran_legacy_radiance["Q"],
                    sasktran_legacy_radiance["U"],
                    sasktran_legacy_radiance["V"],
                ],
                dim="stokes",
            )

            sasktran_legacy_radiance = sasktran_legacy_radiance.drop(
                ["I", "Q", "U", "V"]
            )

        if "wf_brdf" in sasktran_legacy_radiance:
            sasktran_legacy_radiance = sasktran_legacy_radiance.drop("wf_brdf")

        sasktran_legacy_radiance = sasktran_legacy_radiance.rename(
            {"wavelength": "wavelength_nm"}
        )
        sasktran_legacy_radiance = sasktran_legacy_radiance.swap_dims(
            {"wavelength_nm": "spectral_grid"}
        )
        sasktran_legacy_radiance.coords["wavenumber_cminv"] = (
            1e7 / sasktran_legacy_radiance["wavelength_nm"]
        )

        return cls(sasktran_legacy_radiance)

    @classmethod
    def from_sasktran2(
        cls, sasktran2_radiance: xr.Dataset, collapse_scalar_stokes=False
    ) -> SASKTRANRadiance:
        sasktran2_radiance = sasktran2_radiance.rename({"wavelength": "wavelength_nm"})
        sasktran2_radiance = sasktran2_radiance.swap_dims(
            {"wavelength_nm": "spectral_grid"}
        )
        sasktran2_radiance.coords["wavenumber_cminv"] = (
            1e7 / sasktran2_radiance["wavelength_nm"]
        )
        return cls(sasktran2_radiance, collapse_scalar_stokes=collapse_scalar_stokes)

    def _validate_ds(self):
        assert "radiance" in self.data.variables
        assert "wavelength_nm" in self.data.coords
        assert "wavenumber_cminv" in self.data.coords

        assert "stokes" in self.data.radiance.dims
        assert "los" in self.data.radiance.dims
        assert "spectral_grid" in self.data.radiance.dims

        for var in list(self.data.variables):
            if var.startswith("wf_"):
                assert "stokes" in self.data[var].dims
                assert "los" in self.data[var].dims
                assert "spectral_grid" in self.data[var].dims



from __future__ import annotations

import numpy as np
import sasktran as sk

from skretrieval.core import OpticalGeometry
from skretrieval.util import rotation_matrix


class Balloon:
    def __init__(
        self, location, target_altitudes, scan_rate_rad_s, mjd, bore_sight_plane
    ):
        self._location = location
        self._target_altitudes = target_altitudes
        self._scan_rate_rad_s = scan_rate_rad_s
        self._mjd = mjd
        self._bore_sight_plane = bore_sight_plane

    def optical_axis(self, sample_rate_s):
        los_geo = sk.Geodetic()

        low_alt_los = los_geo.from_tangent_altitude(
            self._target_altitudes[0], self._location, self._bore_sight_plane
        )
        high_alt_los = los_geo.from_tangent_altitude(
            self._target_altitudes[1], self._location, self._bore_sight_plane
        )

        rot_axis = np.cross(low_alt_los, high_alt_los)
        rot_axis /= np.linalg.norm(rot_axis)

        current_los = low_alt_los

        theta = self._scan_rate_rad_s * sample_rate_s

        optical_ax = []

        while True:
            # Do motion of balloon
            location = self._location
            local_up = location / np.linalg.norm(location)
            mjd = self._mjd

            optical_ax.append(OpticalGeometry(location, current_los, local_up, mjd))

            los_geo.from_tangent_point(location, current_los)

            if los_geo.altitude > self._target_altitudes[1]:
                break
            current_los = rotation_matrix(rot_axis, theta) @ current_los

        return optical_ax



from __future__ import annotations

import abc

import numpy as np
import sasktran2 as sk2


class Ancillary:
    """
    Ancillary data for the forward model. Typically this is things like temperature/pressure profiles, etc.

    This class is responsible for adding this information to the atmosphere object.
    """

    @abc.abstractmethod
    def add_to_atmosphere(self, atmo: sk2.Atmosphere):
        """
        Adds the ancillary data to the atmosphere object

        Parameters
        ----------
        atmo : sk2.Atmosphere
        """


class GenericAncillary(Ancillary):
    def __init__(
        self,
        altitudes_m: np.array,
        pressure_pa: np.array,
        temperature_k: np.array,
        rayleigh_scattering=True,
    ) -> None:
        """
        A generic ancillary object that can be used to add temperature and pressure profiles to the atmosphere
        as well as include rayleigh scattering.

        Parameters
        ----------
        altitudes_m : np.array
            Altitudes in [m]
        pressure_pa : np.array
            Pressure in [pa]
        temperature_k : np.array
            Temperature in [K]
        rayleigh_scattering : bool, optional
            Whether to add Rayleigh scattering, by default True
        """
        self._altitudes_m = altitudes_m
        self._pressure_pa = pressure_pa
        self._temperature_k = temperature_k
        self._rayleigh_scattering = rayleigh_scattering

    def add_to_atmosphere(self, atmo: sk2.Atmosphere):
        if self._rayleigh_scattering:
            atmo["rayleigh"] = sk2.constituent.Rayleigh()

        atmo.pressure_pa = np.interp(
            atmo.model_geometry.altitudes(), self._altitudes_m, self._pressure_pa
        )
        atmo.temperature_k = np.interp(
            atmo.model_geometry.altitudes(), self._altitudes_m, self._temperature_k
        )


class US76Ancillary(GenericAncillary):
    def __init__(self):
        """
        US76 Standard Atmosphere Ancillary data
        """
        super().__init__(
            sk2.climatology.us76._ALTS,
            sk2.climatology.us76._PRESSURE * 1e4,
            sk2.climatology.us76.celsius_to_kelvin(sk2.climatology.us76._TEMPERATURE_C),
        )



from __future__ import annotations

import fnmatch
from collections.abc import Callable
from dataclasses import dataclass

import numpy as np
import xarray as xr
from scipy import sparse
from scipy.linalg import block_diag
from simpleeval import simple_eval

from skretrieval.core.radianceformat import RadianceGridded


def _resolve_value(expr, variables):
    if isinstance(expr, str):
        expr = expr.replace("$", "")
        return simple_eval(expr, names=variables)
    return expr


@dataclass
class Measurement:
    """
    A dataclass representing the core objects of a measurement vector. This is an internal object
    that is passed around when doing measurement vector transformations
    """

    y: np.array
    K: np.array
    Sy: np.array


class MeasurementVector:
    def __init__(self, fn: Callable, apply_to_filter="*", sample_fn=None):
        """
        A class that represents a measurement vector. This is a callable object that can be used to
        transform L1 data to a measurement vector.

        Parameters
        ----------
        fn : Callable
            Function which takes in L1 data and returns a Measurement object
        apply_to_filter : str, optional
            Only L1 data matching the apply_to_filter will be affected by this measurement vector, by default "*"
        """
        self._sample_fn = sample_fn
        self._fn = fn
        self._filter = apply_to_filter
        self._enabled = True

    @property
    def fn(self):
        return self._fn

    @property
    def filter(self):
        return self._filter

    @property
    def enabled(self):
        return self._enabled

    @enabled.setter
    def enabled(self, value: bool):
        self._enabled = value

    def apply(
        self, l1_data: dict[RadianceGridded], ctxt: dict | None = None
    ) -> Measurement:
        """
        Applies the function to the l1 data, returning back a Measurement object

        Parameters
        ----------
        l1_data : dict[RadianceGridded]

        Returns
        -------
        Measurement
        """
        if not self._enabled:
            return None
        apply_vals = {
            k: d for k, d in l1_data.items() if fnmatch.fnmatch(k, self._filter)
        }
        if len(apply_vals) > 0:
            local_ctxt = ctxt if ctxt is not None else {}
            return self._fn(apply_vals, ctxt=local_ctxt, filter=self._filter)
        return None

    def required_sample_wavelengths(
        self, obs_samples: dict[np.array]
    ) -> dict[np.array]:
        """
        Determines which sample wavelengths are required for this measurement vector

        Default is to just return back all of the observation wavelengths

        Parameters
        ----------
        obs_samples : dict[np.array]

        Returns
        -------
        dict[np.array]
        """
        if self._sample_fn is None:
            return obs_samples
        return self._sample_fn(obs_samples)


def pre_process(l1: dict[RadianceGridded], n: int = 1) -> dict[RadianceGridded]:
    """
    Called before the measurement vector is applied. This function will ensure that the L1 data
    always has the necessary fields for the measurement vector to work.

    Parameters
    ----------
    l1 : dict[RadianceGridded]
    n: int, optional
        Number of elements in the state vector, used to create the dummy Jacobian. Default is 1
        which can be used if the number of elements in the state vector is not important.

    Returns
    -------
    dict[RadianceGridded]
    """
    # Copy and modify the data to always include wf and noise values
    new_l1 = {}

    for key, val in l1.items():
        new_val = val.data.copy(deep=True)

        if "wf" not in new_val:
            new_val["wf"] = xr.zeros_like(
                new_val["radiance"].expand_dims({"x": n}, axis=-1)
            )

        if "radiance_noise" not in new_val:
            new_val["radiance_noise"] = new_val["radiance"] * 1

        new_l1[key] = RadianceGridded(new_val)
    return new_l1


def concat(measurements: list[Measurement]) -> Measurement:
    """
    Concatenates a list of measurements into a single measurement

    Parameters
    ----------
    measurements : list[Measurement]

    Returns
    -------
    Measurement
    """
    if len(measurements) == 0:
        return None
    return Measurement(
        y=np.concatenate([m.y for m in measurements]),
        K=np.vstack([m.K for m in measurements]),
        Sy=sparse.block_diag(
            [sparse.csc_matrix(m.Sy) for m in measurements], format="csc"
        ),
    )


def post_process(measurement: Measurement) -> dict:
    """
    Called after the measurement vector is applied. This function will convert the measurement
    object back into a dictionary for the retrieval to use.

    Parameters
    ----------
    measurement : Measurement

    Returns
    -------
    dict
    """
    # At this stage we have to remove the jacobian if it was a dummy one in the beginning
    res = {"y": measurement.y, "jacobian": measurement.K, "y_error": measurement.Sy}

    if measurement.K.shape[-1] == 0:
        del res["jacobian"]

    return res


def select(l1: dict[RadianceGridded], filter: str = "*", **kwargs) -> Measurement:
    """
    Selects the L1 data that matches the filter and applies the selector stored in kwargs
    to the underlying xarray datasets


    Parameters
    ----------
    l1 : dict[RadianceGridded]
    filter : str, optional
         by default "*"

    Returns
    -------
    Measurement
    """
    measurements = []

    for key, val in l1.items():
        if fnmatch.fnmatch(key, filter):
            selected = val.data.sel(**kwargs)

            measurements.append(
                Measurement(
                    y=selected["radiance"].to_numpy().flatten(),
                    K=selected["wf"].to_numpy().reshape((-1, len(selected["x"]))),
                    Sy=sparse.csc_matrix(
                        sparse.diags(
                            selected["radiance_noise"].to_numpy().flatten() ** 2, 0
                        )
                    ),
                )
            )

    return concat(measurements)


def nearest_selector(l1: dict[RadianceGridded], filter: str = "*", **kwargs) -> dict:
    """
    A special selector that will select the nearest value to the kwargs in the L1 data.
    Returns back another dictionary with the same keys as the input dictionary but with the
    data modified to only contain the nearest values to the kwargs

    Parameters
    ----------
    l1 : dict[RadianceGridded]
    filter : str, optional
        , by default "*"

    Returns
    -------
    dict
    """
    res = {}
    for key, val in l1.items():
        if fnmatch.fnmatch(key, filter):
            res[key] = RadianceGridded(
                val.data.sel(**kwargs, method="nearest").assign_coords(**kwargs)
            )

    return res


def log(measurement: Measurement) -> Measurement:
    """
    Log transform the measurement

    Parameters
    ----------
    measurement : Measurement

    Returns
    -------
    Measurement
    """
    return Measurement(
        y=np.log(measurement.y),
        K=measurement.K / measurement.y[:, np.newaxis],
        Sy=measurement.Sy / np.outer(measurement.y, measurement.y),
    )


def mean(measurement: Measurement) -> Measurement:
    """
    Take the mean of the measurement

    Parameters
    ----------
    measurement : Measurement

    Returns
    -------
    Measurement
    """
    return Measurement(
        y=np.mean(measurement.y),
        K=np.mean(measurement.K, axis=0),
        Sy=np.mean(measurement.Sy.diagonal()),
    )


def multiply(measurement: Measurement, factor: float) -> Measurement:
    """
    Multiply the measurement by a factor

    Parameters
    ----------
    measurement : Measurement
    factor : float

    Returns
    -------
    Measurement
    """
    return Measurement(
        y=measurement.y * factor,
        K=measurement.K * factor,
        Sy=measurement.Sy * factor**2,
    )


def subtract(measurement: Measurement, other: Measurement) -> Measurement:
    """
    Subtract one measurement from another

    Parameters
    ----------
    measurement : Measurement
    other : Measurement

    Returns
    -------
    Measurement
    """
    return Measurement(
        y=measurement.y - other.y,
        K=measurement.K - other.K,
        Sy=(
            measurement.Sy.toarray()
            if sparse.issparse(measurement.Sy)
            else measurement.Sy
        )
        + other.Sy,
    )


def add(measurement: Measurement, other: Measurement) -> Measurement:
    """
    Add two measurements together

    Parameters
    ----------
    measurement : Measurement
    other : Measurement

    Returns
    -------
    Measurement
    """
    return Measurement(
        y=measurement.y + other.y,
        K=measurement.K + other.K,
        Sy=(
            measurement.Sy.toarray()
            if sparse.issparse(measurement.Sy)
            else measurement.Sy
        )
        + other.Sy,
    )


class Triplet(MeasurementVector):
    def __init__(
        self,
        wavelength: list[int],
        weights: list[float],
        altitude_range: list[float],
        normalization_range: list[float],
        normalize=True,
        log_space=True,
        **kwargs,
    ):
        """
        A class that represents a measurement vector that is a weighted combination of log radiances, high altitude normalized

        Note that this measurement vector requires the l1 data to contain the "tangent_altitude" field.

        Both altitude_range and normalization_range can be set through the retrieval context by prefixing the value with a '$'


        Parameters
        ----------
        wavelength : list[int]
            Wavelengths to select
        weights : list[float]
            Weights to apply to the wavelengths
        altitude_range : list[float]
            Altidude range to select
        normalization_range : list[float]
            Altitude range to normalize to
        """
        self._wavelength = wavelength

        def y(l1, ctxt, **kwargs):
            res_altitude_range = [_resolve_value(v, ctxt) for v in altitude_range]
            res_norm_range = [_resolve_value(v, ctxt) for v in normalization_range]

            t_vals = []
            for w, weight in zip(wavelength, weights):
                # Get the useful wavelength data
                if log_space:
                    wavel_data = log(
                        select(
                            nearest_selector(l1, wavelength=w),
                            tangent_altitude=slice(
                                res_altitude_range[0], res_altitude_range[1]
                            ),
                            **kwargs,
                        )
                    )
                else:
                    wavel_data = select(
                        nearest_selector(l1, wavelength=w),
                        tangent_altitude=slice(
                            res_altitude_range[0], res_altitude_range[1]
                        ),
                        **kwargs,
                    )

                # And the normalization value
                norm_vals = mean(
                    log(
                        select(
                            nearest_selector(l1, wavelength=w),
                            tangent_altitude=slice(
                                res_norm_range[0], res_norm_range[1]
                            ),
                            **kwargs,
                        )
                    )
                )

                # The triplet value is the difference of the log radiances subtracted by the normalization multiplied by the weight
                if normalize:
                    t_vals.append(multiply(subtract(wavel_data, norm_vals), weight))
                else:
                    t_vals.append(multiply(wavel_data, weight))
            # Add all of the wavelengths together
            res = t_vals[0]
            for i in range(1, len(t_vals)):
                res = add(res, t_vals[i])

            return res

        super().__init__(y, **kwargs)

    def required_sample_wavelengths(
        self, obs_samples: dict[np.array]
    ) -> dict[np.array]:
        """
        Determines which sample wavelengths are required for this measurement vector

        Default is to just return back all of the observation wavelengths

        Parameters
        ----------
        obs_samples : dict[np.array]

        Returns
        -------
        dict[np.array]
        """
        all_wv = {}

        for key, val in obs_samples.items():
            all_wv[key] = []
            if fnmatch.fnmatch(key, self.filter):
                all_wv[key] = np.array(
                    [val[np.abs(val - w).argmin()] for w in self._wavelength]
                )
        return all_wv



from __future__ import annotations

import numpy as np
import xarray as xr
from sasktran import Geometry, LineOfSight

from skretrieval.core.lineshape import LineShape
from skretrieval.core.radianceformat import RadianceSpectralImage
from skretrieval.legacy.core.sensor import OpticalGeometry, Sensor
from skretrieval.legacy.core.sensor.spectrograph import Spectrograph


class SpectralImager(Sensor):
    """
    Basic ideal instrument for spectral images on a rectangular 2d grid.
    Contains an array of Spectrograph instances to do the calculation for each pixel.

    Parameters
    ----------
    wavelength_nm : np.ndarray
        Array of central wavelengths that the sensor measures.
    spectral_lineshape : LineShape
        Spectral point spread function of the detector. Each pixel is assumed to have the same PSF
    pixel_vert_fov : LineShape
        Vertical angular point spread function of each pixel. Each pixel is assumed to have the same PSF
    pixel_horiz_fov : LineShape
        Horizontal angular point spread function of each pixel. Each pixel is assumed to have the same PSF
    image_horiz_fov : float
        horizontal field of view of the entire detector in radians
    image_vert_fov : float
        vertical field of view of the entire detector in radians
    num_columns: int
        Number of columns in the sensor
    num_rows : int
        Number of rows in the sensor
    """

    def __init__(
        self,
        wavelength_nm: np.ndarray,
        spectral_lineshape: LineShape,
        pixel_vert_fov: LineShape,
        pixel_horiz_fov: LineShape,
        image_horiz_fov: float,
        image_vert_fov: float,
        num_columns: int,
        num_rows: int,
    ):
        self.horizontal_fov = image_horiz_fov
        self.vertical_fov = image_vert_fov
        self._num_columns = num_columns
        self._num_rows = num_rows
        self._wavelenth_nm = wavelength_nm
        self._pixels = self._create_pixels(
            wavelength_nm, spectral_lineshape, pixel_vert_fov, pixel_horiz_fov
        )

    def _create_pixels(
        self,
        wavelength_nm: np.ndarray,
        spectral_lineshape: LineShape,
        pixel_vert_fov: LineShape,
        pixel_horiz_fov: LineShape,
    ) -> list[Spectrograph]:
        pixels = []

        for _i in range(self._num_columns):
            for _j in range(self._num_rows):
                pixels.append(
                    Spectrograph(
                        wavelength_nm=wavelength_nm,
                        pixel_shape=spectral_lineshape,
                        vert_fov=pixel_vert_fov,
                        horiz_fov=pixel_horiz_fov,
                    )
                )

        return pixels

    def model_radiance(
        self,
        optical_geometry: OpticalGeometry,
        model_wavel_nm: np.array,
        model_geometry: Geometry,
        radiance: np.array,
        wf=None,
    ) -> RadianceSpectralImage:
        optical_axes = self.pixel_optical_axes(
            optical_geometry,
            self.horizontal_fov,
            self.vertical_fov,
            self._num_columns,
            self._num_rows,
        )

        model_values = [
            p.model_radiance(oa, model_wavel_nm, model_geometry, radiance, wf=wf)
            for p, oa in zip(self._pixels, optical_axes)
        ]
        return RadianceSpectralImage(
            xr.concat([m.data for m in model_values], dim="los"),
            num_columns=self._num_columns,
        )

    def radiance_format(self) -> type[RadianceSpectralImage]:
        return RadianceSpectralImage

    def measurement_geometry(
        self,
        optical_geometry: OpticalGeometry,
        num_horiz_samples: int | None = None,
        num_vertical_samples: int | None = None,
    ) -> list[LineOfSight]:
        if num_vertical_samples is None:
            num_vertical_samples = self._num_rows
        if num_horiz_samples is None:
            num_horiz_samples = self._num_columns

        optical_axes = self.pixel_optical_axes(
            optical_geometry,
            self.horizontal_fov,
            self.vertical_fov,
            num_horiz_samples,
            num_vertical_samples,
        )
        return [
            LineOfSight(mjd=oa.mjd, observer=oa.observer, look_vector=oa.look_vector)
            for oa in optical_axes
        ]

    def measurement_wavelengths(self) -> np.ndarray:
        return np.unique(np.array([p.measurement_wavelengths() for p in self._pixels]))

    def required_wavelengths(self, res_nm: float) -> np.ndarray:
        return np.unique(
            np.array([p.required_wavelengths(res_nm) for p in self._pixels])
        )

    @staticmethod
    def pixel_optical_axes(
        optical_axis: OpticalGeometry,
        hfov: float,
        vfov: float,
        num_columns: int,
        num_rows: int,
    ) -> list[OpticalGeometry]:
        """
        Get the optical geometry at the center of each pixel in the sensor.

        Parameters
        ----------
        optical_axis : OpticalGeometry
            The optical axis of the center of the sensor
        hfov : float
            horizontal field of view in radians
        vfov : float
            vertical field of view in radians
        num_columns :
            Number of colums in the sensor
        num_rows :
            Number of rows in the sensor

        Returns
        -------
        List[OpticalGeometry]
            The optical geoemetry of each pixel in the sensor as a row-major list.
        """
        pixel_geometry = []
        look = optical_axis.look_vector
        up = optical_axis.local_up
        mjd = optical_axis.mjd
        obs = optical_axis.observer

        dist = 1e6
        htan = np.tan(hfov / 2 * np.pi / 180) * dist
        vtan = np.tan(vfov / 2 * np.pi / 180) * dist
        x = np.linspace(-htan, htan, num_columns)
        y = np.linspace(-vtan, vtan, num_rows)

        nx, ny = np.meshgrid(x, y)
        center = obs + look * dist
        xaxis = np.cross(up, look)
        xaxis /= np.linalg.norm(xaxis)
        yaxis = up
        for (
            x,
            y,
        ) in zip(nx.flatten(), ny.flatten()):
            pos = center + xaxis * x + yaxis * y
            los = pos - obs
            los /= np.linalg.norm(los)
            pixel_yaxis = np.cross(los, up)
            pixel_yaxis /= np.linalg.norm(pixel_yaxis)
            pixel_up = np.cross(pixel_yaxis, los)
            pixel_up /= np.linalg.norm(pixel_up)
            pixel_geometry.append(
                OpticalGeometry(
                    observer=obs, look_vector=los, local_up=pixel_up, mjd=mjd
                )
            )

        return pixel_geometry
